#ifdef WIN32
#include <Winsock2.h>
#pragma comment(lib,"Ws2_32.lib")
#define close closesocket
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include "api.h"

static const char* base64_chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static int is_base64(unsigned char c) 
{
    return (isalnum(c) || (c == '+') || (c == '/'));
}

static const char* base64_encode(unsigned char const* bytes_to_encode, unsigned int in_len) {
	char *ret = NULL;
    int i = 0;
    int j = 0;
    unsigned char char_array_3[3];
    unsigned char char_array_4[4];
	const char* ret_ptr;
   
	ret = (char*)malloc(in_len * 4 + 100);
	if(!ret) return ret;
	ret_ptr = ret;

    while (in_len--) {
        char_array_3[i++] = *(bytes_to_encode++);
        if (i == 3) {
            char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
            char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
            char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
            char_array_4[3] = char_array_3[2] & 0x3f;
           
            for(i = 0; (i <4) ; i++)
				*ret++ = base64_chars[char_array_4[i]];
            i = 0;
        }
    }
   
    if (i)
    {
        for(j = i; j < 3; j++)
            char_array_3[j] = '\0';
       
        char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
        char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
        char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
        char_array_4[3] = char_array_3[2] & 0x3f;
       
        for (j = 0; (j < i + 1); j++)
            *ret++ = base64_chars[char_array_4[j]];
       
        while((i++ < 3))
            *ret++ = '=';
       
    }
	*ret = 0;
   
    return ret_ptr;
}

#ifdef WIN32
int inet_atonx( const char *hostname, struct in_addr *inp )
{
	unsigned long res = 0;
	res = inet_addr( hostname );
	if( res == INADDR_NONE ) {
		// this is not a dotted IP notation string, let's try to resolve a hostname
		struct hostent *he = gethostbyname( hostname );

		if( he == NULL ) {
			return -1;
		}

		memcpy( &inp->S_un, he->h_addr, he->h_length );

		return 0;
	} else {
		inp->S_un.S_addr = res;
		return 0;
	}
}
#else
int inet_atonx(const char *hostname, struct in_addr *inp)
{
	struct hostent * ret;

	if(inet_aton(hostname, inp) != 0) return 0;

	// not dot sep
	ret = gethostbyname(hostname);
	memcpy(inp, ret->h_addr, ret->h_length);
	return 0;
}
#endif

int bc_send_all(int s, char *buf, int len)
{
	int ret;

	while(len > 0)
	{
		ret = send(s, buf, len, 0);
		//ret = write(s, buf, len);
		if(ret < 0) return -1;
		if(ret == 0) return -1;
		buf += ret;
		len -= ret;
	}

	return 0;
}

char *bc_recv_all(int s)
{
	char *buf = (char*)malloc(1024);
	int buf_len = 1024;
	int buf_cur = 0;

	int ret;

	while(1)
	{
		ret = recv(s, buf + buf_cur, buf_len - buf_cur, 0);
		//ret = read(s, buf + buf_cur, buf_len - buf_cur);
		if(ret <= 0) break;
		buf_cur += ret;
		if(buf_cur == buf_len)
		{
			buf = (char*)realloc(buf, buf_len + 10240);
			if(!buf) 
			{
				free(buf);
				return NULL;
			}

			buf_len += 10240;
		}
	}

	buf[buf_cur] = 0;
	return buf;
}

int bc_connect(const char *host, int port)
{
	int s = -1;
	struct sockaddr_in	dst;

#ifdef WIN32
	WORD				wVersionRequested;
	WSADATA				wsaData;
	int					err;
#endif

#ifdef WIN32
	wVersionRequested = MAKEWORD( 2, 2 );
	if(WSAStartup( wVersionRequested, &wsaData ) != 0)
	{
		return -1;
	}

	if ( LOBYTE( wsaData.wVersion ) != 2 ||
		HIBYTE( wsaData.wVersion ) != 2 ) {
			WSACleanup();
			return -1; 
	}
#endif

	s = socket( PF_INET, SOCK_STREAM, 0 );
	if(s < 0) return -1;

	memset( &dst, 0, sizeof( dst ) );
	dst.sin_family	= AF_INET;
	dst.sin_port	= htons( (short)port );
	inet_atonx( host, &dst.sin_addr );

	if( connect( s, (struct sockaddr*)&dst, sizeof( dst ) ) < 0 ) {
		return -1;
	}

	return s;
}

char *bc_post(const char *host, int port, const char *path, const char *d)
{
	int s = bc_connect(host, port);
	char *data = NULL;

	if(s < 0) goto FAIL;
	data = (char *)malloc(10240 + strlen(d));
	if(data == NULL) goto FAIL;

	sprintf(data,
"POST %s HTTP/1.1\r\n"
"Host: %s\r\n"
"User-Agent: Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.9.1.6) Gecko/20091201 Firefox/3.5.6 GTBDFff GTB7.0 (.NET CLR 3.5.30729) WebMoney Advisor\r\n"
"Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8\r\n"
"Accept-Language: en-us,en;q=0.5\r\n"
"Accept-Encoding: deflate\r\n"
"Accept-Charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7\r\n"
"Connection: close\r\n"
"Content-Type: application/x-www-form-urlencoded\r\n"
"Content-Length: %d\r\n"
"\r\n"
"%s",
		path, host, strlen(d), d);

	//printf("==> %s\n", data);

	if(bc_send_all(s, data, strlen(data)) < 0) goto FAIL;
	d = bc_recv_all(s);
	if(!d) goto FAIL;
	
	close(s);
	free(data);
	return (char*)d;

FAIL:
	if(s >= 0) close(s);
	if(data != NULL) free(data);
	return NULL;
}

int bc_load_img(char *buf, int buf_len, const char *file_name)
{
	FILE *fp;
	int file_size;

	if(!file_name) return -1;
	fp = fopen(file_name, "rb");
	if(!fp) return -1;
	fseek(fp, 0, SEEK_END);
	file_size = ftell(fp);
	if(file_size > buf_len)
	{
		fclose(fp);
		return -1;
	}
	fseek(fp, 0, SEEK_SET);
	fread(buf, 1, file_size, fp);
	fclose(fp);
	return file_size;
}

char *bc_encode(const char* img, int img_size)
{
	char *ret = (char*)malloc(img_size * 3 + 100);
	int cur, i;
	char ch;

	if(!ret) return NULL;

	cur = 0;
	for(i = 0;i < img_size;i++)
	{
		ch = img[i];
		if(ch >= 'a' && ch <= 'z' || ch >= 'A' && ch <= 'Z' || ch >= '0' && ch <= '9')
		{
			ret[cur++] = ch;
		}
		else
		{
			cur += sprintf(ret + cur, "%%%02X", (unsigned char)ch);
		}
	}

	ret[cur] = 0;
	return ret;
}

int bc_decode(const char *key, const char *img, int img_size, char *buf, int buf_len, int *task_id)
{
	char *data = (char*)malloc(1024 + img_size * 3);
	const char *img_data = NULL;
	char *ret = NULL, *p, *q;
	int buf_cur = 0;
	const char *new_img_data = NULL;

	if(!data) goto FAIL;

	img_data = base64_encode((const unsigned char*)img, img_size);
	if(!img_data) goto FAIL;

	new_img_data = bc_encode(img_data, strlen(img_data));
	if(!new_img_data) goto FAIL;
	free((void*)img_data);
	img_data = new_img_data;

	sprintf(data, "key=%s&file=%s&gen_task_id=1&base64_code=1", key, img_data);

	ret = bc_post("bypasscaptcha.com", 80, "/upload.php", data);
	if(!ret) goto FAIL;

	//printf("%s\n", ret);

	*task_id = -1;
	if(buf_len > 0) buf[0] = 0;
	p = strtok(ret, "\n");
	while(p)
	{
		if(strncmp(p, "Value ", 6) == 0) {
			q = p + 6;
			buf_cur = 0;
			while(*q && *q != '\r' && *q != '\n' && buf_cur + 1 < buf_len) {
				buf[buf_cur++] = *q++;
			}
			if(buf_cur < buf_len) buf[buf_cur] = 0;
		}
		else if(strncmp(p, "TaskId ", 7) == 0) {
			if(sscanf(p + 7, "%d", task_id) != 1) {
				*task_id = -1;
			}
		}

		p = strtok(NULL, "\n");
	}

	free((void*)img_data);
	free(ret);
	free(data);

	return 0;

FAIL:
	if(img_data) free((void*)img_data);
	if(ret) free(ret);
	if(data) free(data);
	return -1;
}

int bc_check(const char *key, int task_id, int is_correct)
{
	char *ret;
	char buf[1024];

	sprintf(buf, "key=%s&task_id=%d&cv=%d", key, task_id, is_correct);

	ret = bc_post("bypasscaptcha.com", 80, "/check_value.php", buf);

	if(ret == NULL) return -1;
	free(ret);
	return 0;
}
